{# # Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#}
{% extends 'haproxy_proxies.j2' %}
{% set loadbalancer_id = loadbalancer.id %}
{% set usergroup = user_group %}
{% set sock_path = stats_sock %}
{% set connection_limit = loadbalancer.connection_limit %}

{% block proxies %}
{% from 'haproxy_proxies.j2' import frontend_macro as frontend_macro, backend_macro%}
{% for listener in loadbalancer.listeners %}
{{ frontend_macro(constants, listener, loadbalancer.vip_address) }}
{% endfor %}
{% for pool in loadbalancer.pools %}
{{ backend_macro(constants, pool) }}
{% endfor %}
{% endblock proxies %}
