# Copyright (C) 2016 Midokura SARL.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg


lbaas_group = cfg.OptGroup(name='lbaas',
                           title='Neutron LBaaS Options')


lbaas_opts = [
    cfg.StrOpt('default_listener_protocol', default='HTTP',
               help='Default protocol for Neutron LBaaS Listeners'),
    cfg.StrOpt('default_pool_protocol', default='HTTP',
               help='Default protocol for Neutron LBaaS Pools'),
    cfg.StrOpt('default_health_monitor_protocol', default='HTTP',
               help='Default protocol for Neutron LBaaS Health Monitors'),
    cfg.ListOpt('session_persistence_types',
                default=['SOURCE_IP', 'HTTP_COOKIE', 'APP_COOKIE'],
                help='Supported session persistence types for Neutron LBaaS'),
    cfg.BoolOpt('test_with_noop', default=False,
                help='Configures tempest to run with no-op drivers. Ignores '
                     'the operating_status during tests.'),
]
